/*
 * BRLTTY - A background process providing access to the Linux console (when in
 *          text mode) for a blind person using a refreshable braille display.
 *
 * Copyright (C) 1995-2004 by The BRLTTY Team. All rights reserved.
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 *
 * Web Page: http://mielke.cc/brltty/
 *
 * This software is maintained by Dave Mielke <dave@mielke.cc>.
 */

/* spktest.c - Test progrm for the speech synthesizer drivers.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "options.h"
#include "spk.h"
#include "misc.h"

BEGIN_OPTION_TABLE
  {'t', "text-string"   , "string"   , NULL, 0,
   "Text to be spoken."},
  {'D', "data-directory", "directory", NULL, 0,
   "Path to directory for configuration files."},
  {'L', "library-directory", "directory", NULL, 0,
   "Path to directory for loading drivers."},
END_OPTION_TABLE

static const char *opt_textString = NULL;
static const char *opt_dataDirectory = DATA_DIRECTORY;
static const char *opt_libraryDirectory = LIBRARY_DIRECTORY;

static int terminalDescriptor = -1;
static struct termios terminalAttributes;

static const char *fileName;
static int fileLength;
static int fileIndex;

static int
handleOption (const int option) {
  switch (option) {
    default:
      return 0;
    case 't':
      opt_textString = optarg;
      break;
    case 'D':
      opt_dataDirectory = optarg;
      break;
    case 'L':
      opt_libraryDirectory = optarg;
      break;
  }
  return 1;
}

static void *
allocateMemory (size_t size) {
  void *address = malloc(size);
  if (!address) {
    LogError("memory allocation");
    exit(9);
  }
  return address;
}

static void
exitCloseTerminal (void) {
  static const char *trailer = "\r\n";
  write(terminalDescriptor, trailer, strlen(trailer));
  tcsetattr(terminalDescriptor, TCSADRAIN, &terminalAttributes);
  close(terminalDescriptor);
  terminalDescriptor = -1;
}

static int
openTerminal (void) {
  int terminal;
  if ((terminal = open("/dev/tty", O_RDWR|O_NOCTTY)) != -1) {
    if (tcgetattr(terminal, &terminalAttributes) != -1) {
      struct termios attributes = terminalAttributes;
      attributes.c_iflag = IGNBRK | IGNPAR;
      attributes.c_lflag &= ~(ISIG | ICANON | ECHO | ECHOK | ECHONL);
      attributes.c_cc[VMIN] = 0;
      attributes.c_cc[VTIME] = 0;
      if (tcsetattr(terminal, TCSADRAIN, &attributes) != -1) {
        terminalDescriptor = terminal;
        atexit(exitCloseTerminal);
        return 1;
      }
    }

    close(terminal);
  }

  terminalDescriptor = -1;
  return 0;
}

static void
writeStatus (const char *status) {
  if (terminalDescriptor != -1) {
    int percentage = fileLength? (fileIndex*100/fileLength): 0;
    char line[1+0X50+1];
    int length = snprintf(line, sizeof(line), "\r%-7s %10d %3d%% %10d  %44s",
                          status, fileIndex, percentage, fileLength, fileName);
    write(terminalDescriptor, line, length);
  }
}

static int
sayLine (char *line, void *data) {
  sayString(line);

  {
    static time_t oldTime = 0;
    time_t newTime = time(NULL);
    if (newTime != oldTime) {
      oldTime = newTime;
      speech->doTrack();
      fileIndex = speech->getTrack();
      writeStatus("Reading");
    }
  }

  return 1;
}

static void
sayFiles (ECIHand eci, const char *const *paths, int pathCount) {
  static const char *standardInput = "-";
  int pathNumber = 0;

  while (pathNumber < pathCount) {
    const char *path = paths[pathNumber];
    FILE *file;
    if (strcmp(path, standardInput) == 0) {
      fileName = "<standard-input>";
      file = stdin;
    } else {
      fileName = path;
      if (!(file = fopen(path, "r"))) {
        LogPrint(LOG_WARNING, "Cannot open file '%s': %s\n", path, strerror(errno));
        goto nextFile;
      }
    }

    fileIndex = 0;
    if (processLines(file, sayLine, NULL)) {
      int playing = 1;
      int resynthesize = 0;
      const int increment = 200;
      while (eciSpeaking(eci)) {
        fd_set mask;
        struct timeval timeout;
        if (playing) {
          if (resynthesize) {
            resynthesize = 0;
            fseek(file, fileIndex, SEEK_SET);
            eciStop(eci);
            synthesizeFile(eci, file);
            eciSynthesize(eci);
          } else {
            fileIndex = eciGetIndex(eci);
          }
        }
        writeStatus(playing? "Playing": "Paused");
        FD_ZERO(&mask);
        FD_SET(terminalDescriptor, &mask);
        memset(&timeout, 0, sizeof(timeout));
        timeout.tv_sec = 1;
        switch (select(terminalDescriptor+1, &mask, NULL, NULL, &timeout)) {
          case -1:
            if (errno == EINTR) continue;
            systemError("terminal wait");
          case 0:
            continue;
          default: {
            unsigned char character;
            if (read(terminalDescriptor, &character, 1) == 1) {
              switch (character) {
                case 'p':
                  if (playing || !resynthesize) eciPause(eci, playing);
                  playing = !playing;
                  continue;
                case 'q':
                  eciStop(eci);
                  writeStatus("Stopped");
                  return;
                case 'r':
                  fileIndex = 0;
                  resynthesize = 1;
                  continue;
                case 'b':
                  if ((fileIndex -= increment) < 0) fileIndex = 0;
                  resynthesize = 1;
                  continue;
                case 'f':
                  if ((fileIndex += increment) > fileLength) fileIndex = fileLength;
                  resynthesize = 1;
                  continue;
                case 'd':
                  if (pathNumber > 0) {
                    --pathNumber;
                    goto newFile;
                  }
                  break;
                case 'u':
                  if (pathNumber < (pathCount - 1)) {
                    ++pathNumber;
                    goto newFile;
                  }
                  break;
              }
              {
                static const char *alarm = "\a";
                write(terminalDescriptor, alarm, strlen(alarm));
                continue;
              }
            }
            break;
          }
        }
      }
      fileIndex = eciGetIndex(eci);
      writeStatus("End");
    }

  nextFile:
    ++pathNumber;

  newFile:
    if (speech->isSpeaking()) speech->mute();
  }
  writeStatus("Ended");
}

int
main (int argc, char *argv[]) {
  int status;
  const char *driver = NULL;

  processOptions(optionTable, optionCount, handleOption,
                 &argc, &argv, "[driver [parameter=value ...]]");

  if (argc) driver = *argv++, --argc;
  if ((speech = loadSpeechDriver(driver, opt_libraryDirectory))) {
    const char *const *parameterNames = speech->parameters;
    char **parameterSettings;

    if (!parameterNames) {
      static const char *const noNames[] = {NULL};
      parameterNames = noNames;
    }

    {
      const char *const *name = parameterNames;
      unsigned int count;
      char **setting;

      while (*name) ++name;
      count = name - parameterNames;

      parameterSettings = allocateMemory((count + 1) * sizeof(*parameterSettings));
      setting = parameterSettings;
      while (count--) *setting++ = "";
      *setting = NULL;
    }

    while (argc) {
      char *assignment = *argv++;
      int ok = 0;
      char *delimiter = strchr(assignment, '=');
      if (!delimiter) {
        LogPrint(LOG_ERR, "Missing speech driver parameter value: %s", assignment);
      } else if (delimiter == assignment) {
        LogPrint(LOG_ERR, "Missing speech driver parameter name: %s", assignment);
      } else {
        size_t nameLength = delimiter - assignment;
        const char *const *name = parameterNames;
        while (*name) {
          if (strlen(*name) >= nameLength) {
            if (strncasecmp(*name, assignment, nameLength) == 0) {
              parameterSettings[name - parameterNames] = delimiter + 1;
              ok = 1;
              break;
            }
          }
          ++name;
        }
        if (!ok) LogPrint(LOG_ERR, "Invalid speech driver parameter: %s", assignment);
      }
      if (!ok) exit(2);
      --argc;
    }

    if (chdir(opt_dataDirectory) != -1) {
      speech->identify();		/* start-up messages */
      if (speech->open(parameterSettings)) {
        if (opt_textString) {
          sayString(opt_textString);
        } else {
          processLines(stdin, sayLine, NULL);
        }

        speech->close();		/* finish with the display */
        status = 0;
      } else {
        LogPrint(LOG_ERR, "Can't initialize speech driver.");
        status = 5;
      }
    } else {
      LogPrint(LOG_ERR, "Can't change directory to '%s': %s",
               opt_dataDirectory, strerror(errno));
      status = 4;
    }
  } else {
    LogPrint(LOG_ERR, "Can't load speech driver.");
    status = 3;
  }
  return status;
}
